@@ -1,5 +1,5 @@
 /* This file is part of Mailfromd.
-   Copyright (C) 2005-2021 Sergey Poznyakoff
+   Copyright (C) 2005-2022 Sergey Poznyakoff
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -597,7 +597,8 @@ spf_lookup(const char *domain, char **rec)
 	status = txt_lookup(domain, &reply);
 	if (status == dns_success) {
 		int i;
-
+		int found = -1;
+		
 		status = dns_not_found;
 		
 		for (i = 0; i < reply.count; i++) {
@@ -605,11 +606,20 @@ spf_lookup(const char *domain, char **rec)
 					     VSPF1_STR, VSPF1_LEN) == 0
 			    && (reply.data.str[i][VSPF1_LEN] == 0
 				|| mu_isspace(reply.data.str[i][VSPF1_LEN]))) {
-				*rec = mu_strdup(reply.data.str[i]);
-				status = dns_success;
-				break;
+				if (found == -1) {
+					found = i;
+					status = dns_success;
+				} else {
+					found = -1;
+					status = dns_too_many;
+					break;
+				}
 			}
 		}
+
+		if (status == dns_success)
+			*rec = mu_strdup(reply.data.str[found]);
+				
 		dns_reply_free(&reply);
 	}
 	return status;
