@@ -1,5 +1,5 @@
 /* This file is part of Mailfromd.
-   Copyright (C) 2006-2021 Sergey Poznyakoff.
+   Copyright (C) 2006-2022 Sergey Poznyakoff.
    (using my implementation for the GNU tar and rush).
 
    Mailfromd is free software; you can redistribute it and/or modify
@@ -222,6 +222,58 @@ transform_free(struct transform *xform)
     }
 }
 
+static int
+skip_named_class (char const *expr, int i)
+{
+  i++;
+  if (expr[i] == ':')
+    do
+      {
+	for (i++; expr[i] != 0 && expr[i] != ':'; i++)
+	  {
+	    if (expr[i] == '\\')
+	      {
+		i++;
+		if (expr[i] == 0)
+		  return i;
+	      }
+	  }
+	if (expr[i] == 0)
+	  return i;
+	i++;
+	if (expr[i] == ']')
+	  return i + 1;
+      }
+    while (expr[i] != 0);
+
+  return i;
+}
+
+static int
+skip_class (char const *expr, int i)
+{
+  i++;
+  if (expr[i] == ']')
+    i++;
+  while (expr[i] && expr[i] != ']')
+    {
+      switch (expr[i])
+	{
+	case '[':
+	  i = skip_named_class (expr, i);
+	  break;
+
+	case '\\':
+	  if (expr[i+1])
+	    i++;
+	  /* FALLTHROUGH */
+	default:
+	  i++;
+	}
+    }
+  return i;
+}
+
 static int
 parse_transform_expr (struct transform_list *tlist, const char *expr,
 		      int cflags, const char **endp)
@@ -232,18 +284,7 @@ parse_transform_expr (struct transform_list *tlist, const char *expr,
   const char *p;
   struct transform *tf = new_transform (tlist);
   enum transform_type transform_type;
-  static int transition[][8] = {
-    /* '[', ']', ':', '\\', 'a-z', end, dfl, delim */
-    {    1,   0,   0,    6,     0,   8,   0,     7 },
-    {    2,   2,   2,    2,     2,   8,   2,     2 },
-    {    3,   0,   2,    2,     2,   8,   2,     2 },
-    {    2,   2,   4,    6,     2,   8,   2,     2 },
-    {    8,   8,   5,    8,     4,   8,   8,     8 },
-    {    8,   2,   8,    8,     8,   8,   8,     8 },
-    {    0,   0,   0,    0,     0,   8,   0,     0 },
-  };
-  int state;
-
+  
   if (expr[0] != 's')
     {
       last_transform_error.diag = _("invalid transform expression");
@@ -253,56 +294,33 @@ parse_transform_expr (struct transform_list *tlist, const char *expr,
     }
 
   delim = expr[1];
-
+  
   /* Scan regular expression */
-  state = 0;
-  for (i = 2; ; i++)
+  for (i = 2; expr[i]; )
     {
-      int c;
-
-      switch (expr[i])
+      if (expr[i] == '\\')
 	{
-	case '[':
-	  c = 0;
-	  break;
-
-	case ']':
-	  c = 1;
-	  break;
-
-	case ':':
-	  c = 2;
-	  break;
-
-	case '\\':
-	  c = 3;
-	  break;
-
-	case 0:
-	  c = 5;
-	  break;
-
-	default:
-	  if (mu_isalpha (expr[i]))
-	    c = 4;
-	  else if (expr[i] == delim)
-	    c = 7;
-	  else
-	    c = 6;
+	  if (expr[i] == 0)
+	    i += 2;
 	}
-
-      state = transition[state][c];
-      if (state == 7)
+      else if (expr[i] == '[')
+	{
+	  i = skip_class (expr, i);
+	}
+      else if (expr[i] == delim)
 	break;
-      if (state == 8)
-	{
-	  last_transform_error.diag = _("missing 2nd delimiter");
-	  last_transform_error.pos = i;
-	  last_transform_error.arg = expr;
-	  return 1;
-	}
+      else
+	i++;
     }
 
+  if (expr[i] != delim)
+    {
+      last_transform_error.diag = _("missing replacement expression");
+      last_transform_error.pos = i;
+      last_transform_error.arg = expr;
+      return 1;
+    }
+  
   /* Scan replacement expression */
   for (j = i + 1; expr[j] && expr[j] != delim; j++)
     if (expr[j] == '\\' && expr[j+1])
@@ -480,13 +498,7 @@ parse_transform_expr (struct transform_list *tlist, const char *expr,
 	      break;
 
 	    default:
-	      /* Try to be nice */
-	      {
-		char buf[2];
-		buf[0] = '\\';
-		buf[1] = *cur;
-		add_literal_segment (tf, buf, buf + 2);
-	      }
+	      add_char_segment (tf, *cur);
 	      cur++;
 	      break;
 	    }
@@ -782,9 +794,6 @@ run_single_transform (struct transform *tf, struct string_io *iob)
 
 	  disp = rmp[0].rm_eo;
 
-	  if (rmp[0].rm_so)
-	    string_io_append (iob, input, rmp[0].rm_so);
-
 	  nmatches++;
 	  if (tf->match_number && nmatches < tf->match_number)
 	    {
@@ -793,6 +802,9 @@ run_single_transform (struct transform *tf, struct string_io *iob)
 	      continue;
 	    }
 
+	  if (rmp[0].rm_so)
+	    string_io_append (iob, input, rmp[0].rm_so);
+
 	  for (segm = tf->repl_head; segm; segm = segm->next)
 	    {
 	      switch (segm->type)
