@@ -1,5 +1,5 @@
 /* This file is part of Mailfromd.
-   Copyright (C) 2007-2021 Sergey Poznyakoff
+   Copyright (C) 2007-2022 Sergey Poznyakoff
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -521,6 +521,7 @@ struct spf_term_syntax {
 		break;                                            \
 	case dns_not_found:                                       \
 	case dns_failure:                                         \
+	case dns_too_many:					  \
 		return spf_term_nomatch;                          \
 	case dns_temp_failure:                                    \
 		dat->result = spf_temp_error;                     \
@@ -1223,6 +1224,10 @@ spf_check_host_internal(spf_query_t *q, spf_answer_t *a, size_t loopno)
 	case dns_success:
 		break;
 
+	case dns_too_many:
+		SPF_RETURN(spf_perm_error,
+			   "too many SPF records published");
+		
 	case dns_not_found:
 	case dns_failure:
 		SPF_RETURN(spf_none,
